import pytest
import numpy as np

from sklego.common import flatten
from sklego.decomposition import UMAPOutlierDetection
from tests.conftest import general_checks, outlier_checks, select_tests, nonmeta_checks


@pytest.mark.parametrize(
    "test_fn",
    select_tests(
        flatten([general_checks, nonmeta_checks, outlier_checks]),
        exclude=[
            "check_sample_weights_invariance",
            "check_outliers_fit_predict",
            "check_outliers_train",
            "check_fit2d_predict1d",
            "check_methods_subset_invariance",
            "check_fit2d_1sample",
            "check_fit2d_1feature",
            "check_dict_unchanged",
            "check_dont_overwrite_parameters",
            "check_classifier_data_not_an_array"
        ]
    )
)
def test_estimator_checks(test_fn):
    outlier_mod = UMAPOutlierDetection(n_components=2, threshold=0.1)
    test_fn(UMAPOutlierDetection.__name__, outlier_mod)


@pytest.fixture
def dataset():
    np.random.seed(42)
    return np.concatenate([np.random.normal(0, 1, (2000, 10))])


def test_obvious_usecase(dataset):
    mod = UMAPOutlierDetection(n_components=2, threshold=7.5, random_state=42, variant='absolute').fit(dataset)
    assert mod.predict([[10] * 10]) == np.array([-1])
    assert mod.predict([[0.01] * 10]) == np.array([1])



import pytest
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import StandardScaler

from sklego.dummy import RandomRegressor
from sklego.linear_model import DeadZoneRegressor
from sklego.mixture import (
    GMMClassifier,
    BayesianGMMClassifier,
    GMMOutlierDetector,
    BayesianGMMOutlierDetector,
)
from tests.conftest import id_func


@pytest.mark.parametrize(
    "estimator",
    [
        RandomRegressor(strategy="uniform"),
        RandomRegressor(strategy="normal"),
        DeadZoneRegressor(effect="linear", n_iter=100),
        DeadZoneRegressor(effect="quadratic", n_iter=100),
    ],
    ids=id_func,
)
def test_shape_regression(estimator, random_xy_dataset_regr):
    X, y = random_xy_dataset_regr
    assert estimator.fit(X, y).predict(X).shape[0] == y.shape[0]
    pipe = Pipeline(steps=[("scaler", StandardScaler()), ("clf", estimator)])
    assert pipe.fit(X, y).predict(X).shape[0] == y.shape[0]


@pytest.mark.parametrize(
    "estimator",
    [
        GMMClassifier(),
        BayesianGMMClassifier(),
        GMMOutlierDetector(threshold=0.999, method="quantile"),
        GMMOutlierDetector(threshold=2, method="stddev"),
        BayesianGMMOutlierDetector(threshold=0.999, method="quantile"),
        BayesianGMMOutlierDetector(threshold=2, method="stddev"),
    ],
    ids=id_func,
)
def test_shape_classification(estimator, random_xy_dataset_clf):
    X, y = random_xy_dataset_clf
    assert estimator.fit(X, y).predict(X).shape[0] == y.shape[0]
    pipe = Pipeline(steps=[("scaler", StandardScaler()), ("clf", estimator)])
    assert pipe.fit(X, y).predict(X).shape[0] == y.shape[0]



from sklearn import clone
from sklearn.base import BaseEstimator
from sklearn.utils.validation import (
    check_is_fitted,
    check_array,
)

from sklego.common import TrainOnlyTransformerMixin
from deprecated import deprecated


@deprecated(
    version="0.4.2",
    reason="Please use `sklego.preprocessing.OutlierRemovers` instead. "
    "This object will be removed from the meta submodule in version 0.6.0.",
)
class OutlierRemover(TrainOnlyTransformerMixin, BaseEstimator):
    """
    Removes outliers (train-time only) using the supplied removal model.

    :param outlier_detector: must implement `fit` and `predict` methods
    :param refit: If True, fits the estimator during pipeline.fit().
    """

    def __init__(self, outlier_detector, refit=True):
        self.outlier_detector = outlier_detector
        self.refit = refit
        self.estimator_ = None

    def fit(self, X, y=None):
        self.estimator_ = clone(self.outlier_detector)
        if self.refit:
            super().fit(X, y)
            self.estimator_.fit(X, y)
        return self

    def transform_train(self, X):
        check_is_fitted(self, "estimator_")
        predictions = self.estimator_.predict(X)
        check_array(predictions, estimator=self.outlier_detector, ensure_2d=False)
        return X[predictions != -1]



import pytest
import numpy as np

from sklego.linear_model import DeadZoneRegressor
from sklego.testing import check_shape_remains_same_regressor


@pytest.fixture
def dataset():
    np.random.seed(42)
    n = 100
    inputs = np.concatenate([np.ones((n, 1)), np.random.normal(0, 1, (n, 1))], axis=1)
    targets = 3.1 * inputs[:, 0] + 2.0 * inputs[:, 1]
    return inputs, targets


@pytest.fixture(scope="module", params=["linear", "quadratic"])
def mod(request):
    return DeadZoneRegressor(effect=request.param, n_iter=1000)


@pytest.mark.parametrize("test_fn", [check_shape_remains_same_regressor])
def test_deadzone(test_fn):
    regr = DeadZoneRegressor(n_iter=10)
    test_fn(DeadZoneRegressor.__name__, regr)


def test_values_uniform(dataset, mod):
    X, y = dataset
    coefs = mod.fit(X, y).coefs_
    assert coefs[0] == pytest.approx(3.1, abs=0.2)
    assert coefs[1] == pytest.approx(2.0, abs=0.2)



import pytest
import warnings

import pandas as pd
import numpy as np

from sklearn.dummy import DummyClassifier
from sklearn.metrics import accuracy_score
from sklearn.pipeline import make_pipeline
from sklearn.model_selection import GridSearchCV

from sklego.preprocessing import ColumnSelector

from sklego.metrics import subset_score


class DisabledCV:
    """
    DisabledCV is a helper class that can be used for methods and objects requiring crossvalidation
    where you do not want to actually do crossvalidation. For testing purposes we can use this to
    remove any unpredictability
    """

    def __init__(self):
        self.n_splits = 1

    def split(self, X, y, groups=None):
        yield (np.arange(len(X)), np.arange(len(y)))

    def get_n_splits(self, X, y, groups=None):
        return self.n_splits


@pytest.fixture
def slicing_classification_dataset():
    df = pd.DataFrame(
        {
            "x1": [1, 1, 1, 1, 0, 0, 0, 0],
            "x2": [2, 2, 3, 3, 4, 4, 5, 5],
            "y": [0, 1, 1, 1, 0, 0, 0, 1],
        }
    )
    return df[["x1", "x2"]], df["y"]


def test_subset_score_accuracy_pandas(slicing_classification_dataset):
    X, y = slicing_classification_dataset
    model = DummyClassifier(strategy="constant", constant=1).fit(X, y)

    accuracy_x1_0 = subset_score(lambda X, y_true: X["x1"] == 0, accuracy_score)
    accuracy_x1_1 = subset_score(lambda X, y_true: X["x1"] == 1, accuracy_score)
    assert accuracy_x1_0(estimator=model, X=X, y_true=y) == 0.25
    assert accuracy_x1_1(estimator=model, X=X, y_true=y) == 0.75


def test_subset_score_accuracy_numpy(slicing_classification_dataset):
    X, y = slicing_classification_dataset
    X = X.values
    model = DummyClassifier(strategy="constant", constant=1).fit(X, y)

    accuracy_x1_0 = subset_score(lambda X, y_true: X[:, 0] == 0, accuracy_score)
    accuracy_x1_1 = subset_score(lambda X, y_true: X[:, 0] == 1, accuracy_score)
    assert accuracy_x1_0(estimator=model, X=X, y_true=y) == 0.25
    assert accuracy_x1_1(estimator=model, X=X, y_true=y) == 0.75


def test_warning_is_logged_empty_slice(slicing_classification_dataset):
    X, y = slicing_classification_dataset
    model = DummyClassifier(strategy="constant", constant=1).fit(X, y)

    accuracy_x1_0 = subset_score(lambda X, y_true: X["x1"] == 2, accuracy_score)
    with warnings.catch_warnings(record=True) as w:
        # Cause all warnings to always be triggered.
        warnings.simplefilter("always")
        # Trigger a warning.
        accuracy_x1_0(estimator=model, X=X, y_true=y)
        assert issubclass(w[-1].category, RuntimeWarning)


def test_wrong_subset_dimensions_raise_value_error(slicing_classification_dataset):
    X, y = slicing_classification_dataset
    X = X.values
    model = DummyClassifier(strategy="constant", constant=1).fit(X, y)

    accuracy_x1_0 = subset_score(lambda X, y_true: X == 0, accuracy_score)
    with pytest.raises(ValueError):
        accuracy_x1_0(estimator=model, X=X, y_true=y)


def test_subset_score_pipeline(slicing_classification_dataset):
    X, y = slicing_classification_dataset
    model = make_pipeline(
        ColumnSelector("x1"), DummyClassifier(strategy="constant", constant=1)
    ).fit(X, y)

    accuracy_x1_0 = subset_score(lambda X, y_true: X["x1"] == 0, accuracy_score)
    assert accuracy_x1_0(estimator=model, X=X, y_true=y) == 0.25


def test_subset_score_gridsearch(slicing_classification_dataset):
    param_grid = {
        "dummyclassifier__strategy": ["constant"],
        "dummyclassifier__constant": [1],
    }
    pipeline = make_pipeline(DummyClassifier())
    accuracy_x1_0 = subset_score(lambda X, y_true: X["x1"] == 0, accuracy_score)

    cv = DisabledCV()
    search = GridSearchCV(
        estimator=pipeline, param_grid=param_grid, scoring=accuracy_x1_0, cv=cv
    )

    X, y = slicing_classification_dataset
    search.fit(X, y)

    assert search.best_score_ == 0.25



import pytest
import numpy as np
import pandas as pd
from sklearn.linear_model import LinearRegression, LogisticRegression

from sklego.common import flatten
from sklego.meta import RegressionOutlierDetector

from tests.conftest import general_checks, select_tests, outlier_checks


@pytest.mark.parametrize(
    "test_fn",
    select_tests(
        flatten([general_checks, outlier_checks]),
        exclude=['check_fit2d_predict1d', 'check_fit2d_1feature', 'check_outliers_train', 'check_sample_weights_invariance']
        # outliers train wont work because we have two thresholds
    )
)
def test_estimator_checks(test_fn):
    mod = RegressionOutlierDetector(LinearRegression(), column=0)
    test_fn(RegressionOutlierDetector.__name__, mod)


def test_obvious_example():
    # generate random data for illustrative example
    np.random.seed(42)
    X = np.random.normal(0, 1, (100, 1))
    y = 1 + np.sum(X, axis=1).reshape(-1, 1) + np.random.normal(0, 0.2, (100, 1))
    for i in [20, 25, 50, 80]:
        y[i] += 2
    X = np.concatenate([X, y], axis=1)

    # fit and plot
    mod = RegressionOutlierDetector(LinearRegression(), column=1)
    preds = mod.fit(X).predict(X)
    for i in [20, 25, 50, 80]:
        assert preds[i] == -1


def test_obvious_example_pandas():
    # generate random data for illustrative example
    np.random.seed(42)
    x = np.random.normal(0, 1, 100)
    y = 1 + x + np.random.normal(0, 0.2, 100)
    for i in [20, 25, 50, 80]:
        y[i] += 2
    X = pd.DataFrame({'x': x, 'y': y})

    # fit and plot
    mod = RegressionOutlierDetector(LinearRegression(), column='y')
    preds = mod.fit(X).predict(X)
    for i in [20, 25, 50, 80]:
        assert preds[i] == -1


def test_raises_error():
    # generate random data for illustrative example
    np.random.seed(42)
    x = np.random.normal(0, 1, 100)
    y = 1 + x + np.random.normal(0, 0.2, 100)
    for i in [20, 25, 50, 80]:
        y[i] += 2
    X = pd.DataFrame({'x': x, 'y': y})

    with pytest.raises(ValueError):
        mod = RegressionOutlierDetector(LogisticRegression(), column='y')
        mod.fit(X)



import itertools as it
import numpy as np
import pandas as pd
import pytest
from sklearn.linear_model import LinearRegression
from sklearn.preprocessing import StandardScaler, MinMaxScaler
from sklearn.utils import check_X_y

from sklego.common import flatten
from sklego.datasets import load_penguins
from sklego.meta import GroupedTransformer
from tests.conftest import transformer_checks, general_checks, select_tests
from tests.conftest import n_vals, k_vals, np_types


@pytest.mark.parametrize(
    "test_fn",
    select_tests(
        flatten([transformer_checks, general_checks]),
        exclude=[
            # Nonsense checks because we always need at least two columns (group and value)
            "check_fit2d_1feature",
            "check_fit2d_predict1d",
            "check_transformer_data_not_an_array",
        ]
    )
)
def test_estimator_checks(test_fn):
    trf = GroupedTransformer(StandardScaler(), groups=0)
    test_fn(GroupedTransformer.__name__, trf)


@pytest.fixture(
    scope="module", params=[_ for _ in it.product(n_vals, k_vals, np_types)]
)
def dataset_with_single_grouping(request):
    n, k, np_type = request.param
    np.random.seed(42)
    X = np.random.normal(0, 2, (n, k)).astype(np_type)
    y = np.random.normal(0, 2, (n,))

    X, y = check_X_y(X, y)

    # Make sure all groups are present
    groups = np.repeat([0, 1], repeats=(X.shape[0] + 1) // 2)[:X.shape[0], np.newaxis]
    X_with_groups = np.concatenate([groups, X], axis=1)
    grouper = 0  # First column

    return X, y, groups, X_with_groups, grouper


@pytest.fixture(
    scope="module", params=[_ for _ in it.product(n_vals, k_vals, np_types)]
)
def dataset_with_multiple_grouping(request):
    n, k, np_type = request.param
    np.random.seed(42)
    X = np.random.normal(0, 2, (n, k)).astype(np_type)
    y = np.random.normal(0, 2, (n,))

    X, y = check_X_y(X, y)
    # Make sure all groups are present
    groups = np.tile(
        # 4x2 array, repeated until it fits
        np.array([[0, 0], [0, 1], [1, 0], [1, 1]]),
        reps=((X.shape[0] + 3) // 4, 1)
    )[:X.shape[0], :]
    X_with_groups = np.concatenate([groups, X], axis=1)

    grouper = (0, 1)

    return X, y, groups, X_with_groups, grouper


@pytest.fixture(scope="module")
def scaling_range():
    # Some weird interval to make sure we test the right values
    return (13, 42)


@pytest.fixture()
def multiple_obs_fitter():
    from sklearn.base import BaseEstimator, TransformerMixin

    class MultipleObsFitter(BaseEstimator, TransformerMixin):
        """A transformer that needs more than 1 value to fit"""
        def fit(self, X, y=None):
            if len(X) <= 1:
                raise ValueError("Need more than 1 value to fit")

            return self

        def transform(X):
            return X

    return MultipleObsFitter()


@pytest.fixture(scope="module")
def penguins_df():
    df = load_penguins(as_frame=True).dropna()
    X = df.drop(columns='species')

    return X


@pytest.fixture(scope="module")
def penguins(penguins_df):
    return penguins_df.values


def test_all_groups_scaled(dataset_with_single_grouping, scaling_range):
    X, y, groups, X_with_groups, grouper = dataset_with_single_grouping

    trf = MinMaxScaler(scaling_range)
    transformer = GroupedTransformer(trf, groups=grouper)
    transformed = transformer.fit(X_with_groups, y).transform(X_with_groups)

    df_with_groups = pd.concat([pd.Series(groups.flatten(), name="G"), pd.DataFrame(transformed)], axis=1)

    assert np.allclose(df_with_groups.groupby("G").min(), scaling_range[0])
    assert np.allclose(df_with_groups.groupby("G").max(), scaling_range[1])


def test_group_correlation_minmaxscaler(dataset_with_single_grouping, scaling_range):
    X, y, groups, X_with_groups, grouper = dataset_with_single_grouping

    trf = MinMaxScaler(scaling_range)
    transformer = GroupedTransformer(trf, groups=grouper)
    transformed = transformer.fit(X_with_groups, y).transform(X_with_groups)

    # For each column, check that all grouped correlations are 1 (because MinMaxScaler scales linear)
    for col in range(X.shape[1]):
        assert (
            pd.concat([
                pd.Series(groups.flatten(), name="group"),
                pd.Series(X[:, col], name="original"),
                pd.Series(transformed[:, col], name="transformed"),
            ], axis=1)
            .groupby("group")
            .corr()
            .pipe(np.allclose, 1)
        )


def test_set_params():
    trf = StandardScaler(with_std=False)
    transformer = GroupedTransformer(trf, groups=0)

    transformer.set_params(transformer__with_std=True)
    assert trf.with_std


def test_get_params():
    trf = StandardScaler(with_std=False)
    transformer = GroupedTransformer(trf, groups=0)

    assert transformer.get_params() == {
        "transformer__with_mean": True,
        "transformer__with_std": False,
        "transformer__copy": True,
        "transformer": trf,
        "groups": 0,
        "use_global_model": True,
    }


def test_non_transformer(dataset_with_single_grouping):
    X, y, _, _, grouper = dataset_with_single_grouping

    # This is not a transformer
    trf = LinearRegression()
    transformer = GroupedTransformer(trf, groups=grouper)

    with pytest.raises(ValueError):
        transformer.fit(X, y)


def test_multiple_grouping_columns(dataset_with_multiple_grouping, scaling_range):
    X, y, groups, X_with_groups, grouper = dataset_with_multiple_grouping

    trf = MinMaxScaler(scaling_range)
    transformer = GroupedTransformer(trf, groups=grouper)
    transformed = transformer.fit(X_with_groups, y).transform(X_with_groups)

    df_with_groups = pd.concat([
        pd.DataFrame(groups, columns=["A", "B"]),
        pd.DataFrame(transformed)
    ], axis=1)

    assert np.allclose(df_with_groups.groupby(["A", "B"]).min(), scaling_range[0])

    # If a group has a single element, it defaults to min, so check wether all maxes are one of the bounds
    maxes = df_with_groups.groupby(["A", "B"]).max()
    assert np.all(
        np.isclose(maxes, scaling_range[1]) | np.isclose(maxes, scaling_range[0])
        # We have at least some groups larger than 1, so there we should find the max of the range
    ) and np.any(np.isclose(maxes, scaling_range[1]))


def test_missing_groups_transform_global(dataset_with_single_grouping, scaling_range):
    X, y, groups, X_with_groups, grouper = dataset_with_single_grouping

    trf = MinMaxScaler(scaling_range)
    transformer = GroupedTransformer(trf, groups=grouper)
    transformer.fit(X_with_groups, y)

    # Array with 2 rows, first column a new group. Remaining top are out of range so should be the range
    X_test = np.concatenate([
        np.array([[3], [3]]), np.stack([X.min(axis=0), X.max(axis=0)], axis=0)
    ], axis=1)

    transformed = transformer.transform(X_test)

    # Top row should all be equal to the small value of the range, bottom the other
    assert np.allclose(transformed[0, :], scaling_range[0])
    assert np.allclose(transformed[1, :], scaling_range[1])


def test_missing_groups_transform_noglobal(dataset_with_single_grouping, scaling_range):
    X, y, groups, X_with_groups, grouper = dataset_with_single_grouping

    trf = MinMaxScaler(scaling_range)
    transformer = GroupedTransformer(trf, groups=grouper, use_global_model=False)
    transformer.fit(X_with_groups, y)

    # Array with 2 rows, first column a new group. Remaining top are out of range so should be the range
    X_test = np.concatenate([
        np.array([[3], [3]]), np.stack([X.min(axis=0) - 1, X.max(axis=0) + 1], axis=0)
    ], axis=1)

    with pytest.raises(ValueError):
        transformer.transform(X_test)


def test_exception_in_group(multiple_obs_fitter):
    X = np.array([
        [1, 2],
        [1, 0],
        [2, 1],
    ])

    # Only works on groups greater than 1, so will raise an error in group 2
    transformer = GroupedTransformer(multiple_obs_fitter, groups=0, use_global_model=False)

    with pytest.raises(ValueError) as e:
        transformer.fit(X)

        assert "group 2" in str(e)


def test_array_with_strings():
    X = np.array([
        ("group0", 2),
        ("group0", 0),
        ("group1", 1),
        ("group1", 3),
    ], dtype='object')

    trf = MinMaxScaler()
    transformer = GroupedTransformer(trf, groups=0, use_global_model=False)
    transformer.fit_transform(X)


def test_df(penguins_df):
    meta = GroupedTransformer(StandardScaler(), groups=["island", "sex"])

    transformed = meta.fit_transform(penguins_df)

    # 2 columns for grouping not in the result
    assert transformed.shape == (penguins_df.shape[0], penguins_df.shape[1] - 2)


def test_df_missing_group(penguins_df):
    meta = GroupedTransformer(StandardScaler(), groups=["island", "sex"])

    # Otherwise the fixture is changed
    X = penguins_df.copy()
    X.loc[0, "island"] = None

    with pytest.raises(ValueError):
        meta.fit_transform(X)


def test_array_with_multiple_string_cols(penguins):
    X = penguins

    meta = GroupedTransformer(StandardScaler(), groups=[0, -1])

    transformed = meta.fit_transform(X)

    # 2 columns for grouping not in the result
    assert transformed.shape == (X.shape[0], X.shape[1] - 2)


def test_grouping_column_not_in_array(penguins):
    X = penguins

    meta = GroupedTransformer(StandardScaler(), groups=[0, 5])

    # This should raise ValueError
    with pytest.raises(ValueError):
        meta.fit_transform(X[:, :3])


def test_grouping_column_not_in_df(penguins_df):
    meta = GroupedTransformer(StandardScaler(), groups=["island", "unexisting_column"])

    # This should raise ValueError
    with pytest.raises(ValueError):
        meta.fit_transform(penguins_df)


def test_no_grouping(penguins_df):
    penguins_numeric = penguins_df[["bill_length_mm", "bill_depth_mm", "flipper_length_mm", "body_mass_g"]]

    meta = GroupedTransformer(StandardScaler(), groups=None)
    nonmeta = StandardScaler()

    assert (meta.fit_transform(penguins_numeric) == nonmeta.fit_transform(penguins_numeric)).all()


def test_with_y(penguins_df):
    X = penguins_df.drop(columns=["sex"])
    y = penguins_df["sex"]

    meta = GroupedTransformer(StandardScaler(), groups="island")

    # This should work fine
    transformed = meta.fit_transform(X, y)

    # 1 column for grouping not in the result
    assert transformed.shape == (X.shape[0], X.shape[1] - 1)
